/* FUNCTIONS.C 
 * Copyright 1998-2001 Stephen C. Grubb  (ploticus.sourceforge.net) .
 * This code is covered under the GNU General Public License (GPL);
 * see the file ./Copyright for details. */

#include <stdio.h>
#define MAXARGS 64

#define NUMBER 0
#define ALPHA 1

extern char *TDH_dat, *TDH_recid;
#define err(a,b,c) 		TDH_err(a,b,c)

/* the following is used by $tmpfilename.. */
extern char TDH_tmpdir[];
extern char TDH_scriptdir[];
#if TDH_DB == 1
extern char DB_datapath[];
#endif

static int evalflag;
static char Sep[5] = ","; /* for args that are commalists */
static int eval_function();

char *GL_getok();

/* =============================================== */
/* FUNCTION_CALL - parse function name, and any arguments, then call
   function evaluator.  V is returned with the result value. */

TDH_function_call( v, typ, eval )
char *v;
int *typ;  /* sent back.. describes basic type of result.. either NUMBER or ALPHA */
int eval;  /* 1 if unevaluated @variables possible, 0 otherwise */
{
int i, ix, len, status;
char name[80];
char tok[256];
char argbuf[ 1024 ];
int nargs;
int alen;
char *arg[MAXARGS];

evalflag = eval;

/* get function name */
ix = 0;
len = strlen( v );
GL_getchunk( name, v, &ix, " (" );
if( name[0] == '\0' ) {
	err( 1601, "$function expected, nothing found", "" );
	return( 1 );
	}
/* get function parameters, parsing them into arg pointer array */
alen = 0;
for( i = 0; i < MAXARGS; i++ ) {
	if( ix >= len ) break;
	GL_getchunk( tok, v, &ix, " ,()" );
	arg[i] = &argbuf[ alen ];
	if( evalflag ) TDH_value_subst( &argbuf[ alen ], tok, TDH_dat, TDH_recid, 0, 0 );
	else strcpy( &argbuf[ alen ], tok );
	alen += strlen( &argbuf[ alen ] ) + 1;
	}
nargs = i;

nargs--; /* above parsing always picks up an extra arg for closing paren */

status = eval_function( name, arg, nargs, v, typ );
return( status );
}

/* ======================================== */
/* EVAL_FUNCTION - implement general purpose condex functions. 

	Return 0 if ok, 1 if named function doesn't exist or some other error. 
	Functions which produce a result which should always be taken as
   	alpha rather than number should append a space to the end of the result. 
*/
static int
eval_function( name, arg, nargs, result, typ )
char *name; 	/* function name */
char *arg[];    /* arg list */
int nargs;	/* number of args */
char *result;	/* the result value gets copied into this string */
int *typ;	/* basic type of result */
{
int stat;
double accum;
char curop;
double atof();
int i, j;
char tok[256];
char *s;

/* printf( "[function call name=%s nargs=%d]", name, nargs );
 * for( i = 0; i < nargs; i++ ) printf( "[%s]", arg[i] );
 * return( 0 );
 */

*typ = ALPHA; /* fallback */

/* ------------------------------------------------------ */
/* Built-in functions                                     */
/* ------------------------------------------------------ */

/* -----*/
/* LEN  */
/* -----*/

/* ----------------- */
/* $len(s) - return length of argument */
if( strcmp( name, "$len" )==0 ) {
	sprintf( result, "%d", strlen( arg[0] ) );
	*typ = NUMBER;
	return( 0 );
	}

/* ------- */
/* CUSTOM  */
/* ------- */

/* ----------------- */
/* user-defined functions */
else if( name[0] == '$' && name[1] == '$' ) {
	stat = custom_function( &name[2], arg, nargs, result, typ );
	if( stat != 0 ) err( 1602, "error on custom function", name );
	return( stat );
	}

/* -------------- */
/* SQL and SHELL */
/* -------------- */
#ifdef TDH_DB
else if( strncmp( name, "$sql", 4 )==0 ) 
	return( TDH_dbfunctions( name, arg, nargs, result, typ, TDH_dat, TDH_recid ) );
#endif

else if( strncmp( name, "$shell", 6 )==0 ) 
	return( TDH_shfunctions( name, arg, nargs, result, typ, TDH_dat, TDH_recid ) );

/* ------- */
/* NUMERIC */
/* ------- */

/* ------------ */
/* $arith() - L to R arithmetic expression evaluator, e.g.  4+3*2-1 (unary plus/minus allowed) */
/* $arithl() - same as above except lazy (non-numeric arguments are silently allowed and treated as 0) */
else if( GL_smember( name, "$arith $arithl" )) {
	i = 0;
	accum = 0.0;
	curop = '+';
	s = arg[0];
	s[ strlen( s ) + 1 ] = '\0'; /* add extra null so loop terminates correctly */
	if( s[0] == '-' ) tok[0] = '-'; /* check for unary minus on first arg */
	else tok[0] = '+';
	while( 1 ) {
		GL_getchunk( &tok[1], s, &i, "+-/*%" ); /* tok[1] because sign is always in tok[0] */
		if( strlen( &tok[1] ) < 1 ) break;
		if( ! GL_goodnum( tok, &stat ) && name[6] != 'l' )  
			{ err( 1603, "arith: bad value", tok ); return( 1 ); }
		if( curop == '+' ) accum = accum + atof( tok );
		else if( curop == '-' ) accum = accum - atof( tok );
		else if( curop == '/' ) {
			if( atof( tok ) == 0.0 ) {  /* added scg 8/4/01 */
				err( 1605, "arith: divide by zero", "" );
				return( 1 );
				}
			accum = accum / atof( tok );
			}
		else if( curop == '*' ) accum = accum * atof( tok );
		else if( curop == '%' ) accum = (int)(accum) % (int)(atof( tok ));
		else { err( 1604, "arith:bad operator", "" ); return( 1 ); }
		curop = s[i];
		if( s[i+1] == '-' ) tok[0] = '-'; /* check for unary minus on next operand */
		else tok[0] = '+';
		i++;
		}
	sprintf( result, "%g", accum );
	*typ = NUMBER;
	return( 0 );
	}

/* ----------- */
/* $inr(n,low,hi) - see if n is in range low to high */
else if( strcmp( name, "$inr" )==0 ) {
	i = 0;
	if( !GL_goodnum( arg[0], &i ) ) strcpy( result, "0" );
	else if( atof( arg[0] ) >= atof( arg[1] ) && atof( arg[0] ) <= atof( arg[2] ) ) strcpy( result, "1" );
	else strcpy( result, "0" );
	*typ = NUMBER;
	return( 0 );
	}

/* ----------- */
/* $formatfloat(n,format) */
else if( strcmp( name, "$formatfloat" )==0 ) {
	sprintf( result, arg[1], atof( arg[0] ) );
	*typ = ALPHA;  /* apparently */
	return( 0 );
	}

/* -------------- */
/* $autoround(val,d) - round val to reasonable precision */
else if( strcmp( name, "$autoround" )==0 ) {
	strcpy( result, arg[0] );
	GL_autoround( result, atoi( arg[1] ) );
	*typ = NUMBER;
	return( 0 );
	}

/* ----------- */
/* $isnumber(s) - result is 1 if argument is a valid number, 0 if not. */
else if( strcmp( name, "$isnumber" )==0 ) {
	if(  arg[0][0] == '\0' ) {
		sprintf( result, "0" ) ;
		return( 0 );
		}
	stat = GL_goodnum( arg[0], &i );
	sprintf( result, "%d", stat );
	*typ = NUMBER;
	return( 0 );
	}

/* ------------ */
/* $math(what,f) - return math function of f */
else if( strcmp( name, "$math" )==0 ) {
	double fabs(), sqrt();
	if( strcmp( arg[0], "abs" )==0 ) sprintf( result, "%g", fabs( atof( arg[1] ) ) );
	/* else if( strcmp( arg[0], "sqrt" )==0 ) sprintf( result, "%g", sqrt( atof( arg[1] ) ) ); */
	/* others may be added here... */
	*typ = NUMBER;
	return( 0 );
	}

/* ------------- */
/* $random() - return a random number between 0 and 1 */
else if( strcmp( name, "$random" )==0 ) {
	double GL_rand();
	sprintf( result, "%g", GL_rand() );
	*typ = NUMBER;
	return( 0 );
	}

/* ------- */
/* STRING */
/* ------- */

/* ------------ */
/* $strlen(s) - result is length of string s */
else if( strcmp( name, "$strlen" )==0 ) {
	sprintf( result, "%d", strlen( arg[0] ) );
	*typ = NUMBER;
	return( 0 );
	}

/* ------------ */
/* $strcat(s,t) - concatenate s and t */
else if( strcmp( name, "$strcat" )==0 ) {
	sprintf( result, "%s%s", arg[0], arg[1] );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------ */
/* $substring(s,from,nchar) - result is a substring, e.g. $substring(abcde,3,99) --> cde */
else if( strcmp( name, "$substring" )==0 ) {
	GL_substring( result, arg[0], atoi( arg[1] ), atoi( arg[2] ) );
	*typ = ALPHA;
	return( 0 );
	}

/* ----------- */
/* $change(s1,s2,line) - in line, substitute every occurance of s1 with s2. */
else if( strcmp( name, "$change" )==0 ) {
	stat = GL_substitute( arg[0], arg[1], arg[2] ); /* modifies argbuf */
	sprintf( result, "%s", arg[2] );
	*typ = ALPHA;
	return( 0 );
	}

/* ----------- */
/* $char(n,s) - get the nth character of string s */
else if( strcmp( name, "$char" )==0 ) {
	int n;
	n = atoi( arg[0] );
	if( n < 1 || n > (strlen( arg[1] )-1) ) return( -1 );
	sprintf( result, "%c", arg[1][ n-1 ] );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------ */
/* $ntoken( n, s ) - return the nth whitespace-delimited token in s */
else if( strcmp( name, "$ntoken" )==0 ) {
	int ix, n;
	n = atoi( arg[0] );
	ix = 0;
	for( i = 0; i < n; i++ ) strcpy( result, GL_getok( arg[1], &ix ) );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------- */
/* $extractnum( s ) - extract the first numeric entity embedded anywhere in s */
else if( strcmp( name, "$extractnum" )==0 ) {
	int len;
	for( i = 0, len = strlen( arg[0] ); i < len; i++ ) if( isdigit( arg[0][i] ) ) break;
	if( i == len ) strcpy( result, "" );
	else sprintf( result, "%g", atof( &arg[0][i] ) );
	return( 0 );
	}

/* ------------ */
/* $lowerc(string) - result is the lower-case equivalent of string */
else if( strcmp( name, "$lowerc" )==0 ) {
	int slen;
	strcpy( result, arg[0] );
	for( i = 0, slen = strlen( result ); i < slen; i++ ) result[i] = tolower( result[i] );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------ */
/* $upperc(string) - result is the lower-case equivalent of string */
else if( strcmp( name, "$upperc" )==0 ) {
	int slen;
	strcpy( result, arg[0] );
	for( i = 0, slen = strlen( result ); i < slen; i++ ) result[i] = toupper( result[i] );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------- */
/* $contains(clist,s) - if string s contains any of chars in clist, result is position
 *	(first=1) of first occurance.  Result is 0 if none found. */
else if( strcmp( name, "$contains" )==0 ) {
	sprintf( result, "%d", GL_contains( arg[0], arg[1] ) );
	*typ = NUMBER;
	return( 0 );
	}

/* ------------- */
/* $changechars(clist,s,newchar) - if string s contains any of chars in clist, change that
 *	character to newchar */
else if( strcmp( name, "$changechars" )==0 ) {
	strcpy( tok, arg[1] );
	GL_changechars( arg[0], tok, arg[2] );
	sprintf( result, "%s", tok );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------- */
/* $deletechars(clist,s) - if string s contains any of chars in clist, delete that character */
else if( strcmp( name, "$deletechars" )==0 ) {
	strcpy( tok, arg[1] );
	GL_deletechars( arg[0], tok );
	sprintf( result, "%s", tok );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------- */
/* $fuzzymatch(s,t,tightness) */
else if( strcmp( name, "$fuzzymatch" )==0 ) {
	stat = GL_fuzzymatch( arg[0], arg[1], strlen(arg[1]), atoi( arg[2] ) );
	sprintf( result, "%d", stat );
	*typ = NUMBER;
	return( 0 );
	}


/* ------------ */
/* COMMALIST   */
/* ----------- */

/* --------------- */
/* $count(str,list) - count the number of times 'str' appears in 'list'.  List is
	a comma-delimited list.  Example: $count(hello,aba,gabba,jabba) would yield 0 */
/* if str is '*', result is the number of members in list */
else if( strcmp( name, "$count" )==0 ) {
	int n, j, len;
	n = 0;
	for( j = 1; j < nargs; j++ ) {  /* allow comma lists in several args.. */
		s = arg[j];
		i = 0;
		len = strlen( s );
		while( i < len ) {
			GL_getseg( tok, s, &i, Sep );   /* parse list */
			/* if( strlen( tok ) < 1 ) break; */
			if( strcmp( arg[0], "*" )==0 ) n++;
			else if( strcmp( tok, arg[0] )==0 ) n++;
			}
		}

	sprintf( result, "%d", n );
	*typ = NUMBER;
	return( 0 );
	}

/* --------------- */
/* $nmember(n,list) - result is the nth member of the list.  List is a comma-delimited list. */
else if( strcmp( name, "$nmember" )==0 ) {
	int j, n;
	n = atoi( arg[0] );
	s = arg[1];
	j = 0; i = 0;
	while( 1 ) {
		j++;
		GL_getseg( tok, s, &i, Sep );  /* parse list */
		if( tok[0] == '\0' ) { strcpy( result, "" ); break; }
		else if( j == n ) { sprintf( result, "%s", tok ); break; }
		}
	*typ = ALPHA;
	return( 0 );
	}

/* --------------- */
/* $addmember(newmem,list) - append newmem to the end of comma-delimited list. */
else if( strcmp( name, "$addmember" )==0 ) {
	strcpy( result, arg[1] );
	GL_addmember( arg[0], result ); 
	*typ = ALPHA;
	return( 0 );
	}

/* -------------- */
/* $commonmembers( list1, list2 ) - return # of members in common between 2 lists */
else if( strcmp( name, "$commonmembers" )==0 ) {
	int mode;
	if( strcmp( arg[2], "count" )==0 ) mode = 0;
	else mode = 1;
	sprintf( result, "%d", GL_commonmembers( arg[0], arg[1], mode ) );
	*typ = NUMBER;
	return( 0 );
	}

/* -------------- */
/* $makelist( s ) - convert white-space separated list to commalist */
else if( strcmp( name, "$makelist" )==0 ) {
	int ix;
	strcpy( result, "" );
	ix = 0;
	while( 1 ) {
		strcpy( tok, GL_getok( arg[0], &ix ) );
		if( tok[0] == '\0' ) break;
		if( tok[ strlen( tok ) - 1 ] == ',' ) tok[ strlen( tok ) - 1 ] = '\0'; 
		GL_addmember( tok, result );
		}
	*typ = ALPHA;
	return( 0 );
	}


/* ------- */
/* MISC    */
/* ------- */

/* ------------- */
/* $uniquename() - result is a unique name based on date, time, and pid() */
else if( strcmp( name, "$uniquename" )==0 ) {
	GL_make_unique_string( result, 0 );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------------ */
/* $tmpfilename(tag) - return a tmp file name using tmpdir from config file */
else if( strcmp( name, "$tmpfilename" )==0 ) {
	GL_make_unique_string( tok, 0 );
	if( arg[0][0] != '\0' ) strcat( arg[0], "." );
	sprintf( result, "%s/%s%s", TDH_tmpdir, arg[0], tok );
	*typ = ALPHA;
	return( 0 );
	}

/* ------------------ */
/* $getenv(envvarname) - get the contents of the named shell environment variable */
else if( strcmp( name, "$getenv" )==0 ) {
	char *getenv();
	char *s;
	int len;
	s = getenv( arg[0] );
	if( s == NULL ) strcpy( result, "" );
	else strcpy( result, s );
	/* for( i = 0, len = strlen( result ); i < len; i++ ) if( result[i] == ' ') result[i] = '_'; */
	*typ = ALPHA;
	return( 0 );
	}

/* ------------------ */
/* $fileexists( dir, name ) - return 1 if the requested file can be opened, 0 if not.  dir
	is a symbol indicating what directory name is relative to (see docs) */
else if( strcmp( name, "$fileexists" )==0 ) {
	FILE *testfp;
	if( strcmp( arg[0], "/" )==0 ) sprintf( tok, "%s", arg[1] );
	else if( strcmp( arg[0], "scriptdir" )==0 ) sprintf( tok, "%s/%s", TDH_scriptdir, arg[1] );
#if TDH_DB == 1
	else if( strcmp( arg[0], "datadir" )==0 ) sprintf( tok, "%s/%s", DB_datapath, arg[1] );
#endif
	else if( strcmp( arg[0], "tmpdir" )==0 ) sprintf( tok, "%s/%s", TDH_tmpdir, arg[1] );
	testfp = fopen( tok, "r" );
	if( testfp == NULL ) strcpy( result, "0" );
	else	{
		strcpy( result, "1" );
		fclose( testfp );
		}
	*typ = NUMBER;
	return( 0 );
	}



/* ------------------ */
/* $ref( varname ) - return the contents of variable varname */
else if( strcmp( name, "$ref" )==0 ) {
	TDH_getvalue( result, arg[0], TDH_dat, TDH_recid );
	return( 0 );
	}

/* ------------------ */
/* $def( varname ) - return 1 if variable has been defined, 0 if not */
else if( strcmp( name, "$def" )==0 ) {
	if( TDH_getvalue( result, arg[0], TDH_dat, TDH_recid ) == 0 ) sprintf( result, "1" );
	else sprintf( result, "0" );
	return( 0 );
	}

/* ------------------------ */
/* $htmlquote( var1, .. varN ) - convert embedded quotes (") to &quot; for all variables given */
else if( strcmp( name, "$htmlquote" )==0 ) {
	for( i = 0; i < nargs; i++ ) {
        	stat = TDH_getvalue( tok, arg[i], TDH_dat, TDH_recid );
        	GL_substitute( "\"", "&quot;", tok );
        	stat = TDH_setvalue( arg[i], tok, TDH_dat, TDH_recid );
		}
	sprintf( result, "0" );
	return( 0 );
	}

/* ------------------------ */
/* $htmldisplay( var1, .. varN ) - convert embedded HTML problem chars (<,>,&) to their html esc sequences
	for all variables given */
else if( strcmp( name, "$htmldisplay" )==0 ) {
	char *strchr();
	for( i = 0; i < nargs; i++ ) {
        	stat = TDH_getvalue( tok, arg[i], TDH_dat, TDH_recid );
		if( strchr( tok, '&' ) != NULL ) GL_substitute( "&", "&amp;", tok );
		if( strchr( tok, '<' ) != NULL ) GL_substitute( "<", "&lt;", tok );
		if( strchr( tok, '>' ) != NULL ) GL_substitute( ">", "&gt;", tok );
        	stat = TDH_setvalue( arg[i], tok, TDH_dat, TDH_recid );
		}
	sprintf( result, "0" );
	return( 0 );
	}


/* ------------------------------- */
/* DB, SHELL, DATE, TIME, CHECKSUM */
/* ------------------------------- */
stat = DT_datefunctions( name, arg, nargs, result, typ );
if( stat != 2 ) return( stat );
stat = DT_timefunctions( name, arg, nargs, result, typ );
if( stat != 2 ) return( stat );
stat = GL_checksum_functions( name, arg, nargs, result, typ );
if( stat != 2 ) return( stat );


/* ------- */
/* ELSE... */
/* ------- */

/* otherwise... try a user-defined function */

stat = custom_function( &name[1], arg, nargs, result, typ );
if( stat != 0 ) err( 1610, "unrecognized function", name );
return( stat );

}

/* ==================== */
TDH_function_listsep( sepchar )
char sepchar;
{
Sep[0] = sepchar;
Sep[1] = '\0';
return( 0 );
}
